@@ -99,6 +99,10 @@ function LanguageTree.new(source, lang, opts)
   ---@type LanguageTreeOpts
   opts = opts or {}
 
+  if source == 0 then
+    source = vim.api.nvim_get_current_buf()
+  end
+
   local injections = opts.injections or {}
   local self = setmetatable({
     _source = source,
@@ -262,7 +266,7 @@ function LanguageTree:parse()
           tcall(self._parser.parse, self._parser, self._trees[i], self._source, true)
 
         -- Pass ranges if this is an initial parse
-        local cb_changes = self._trees[i] and tree_changes or ranges
+        local cb_changes = self._trees[i] and tree_changes or tree:included_ranges(true)
 
         self:_do_callback('changedtree', cb_changes, tree)
         self._trees[i] = tree
@@ -308,9 +312,9 @@ function LanguageTree:parse()
     query_time = query_time,
   })
 
-  self:for_each_child(function(child)
+  for _, child in pairs(self._children) do
     child:parse()
-  end)
+  end
 
   self._valid = true
 
@@ -517,11 +521,13 @@ end
 ---@param node TSNode
 ---@param source string|integer
 ---@param metadata TSMetadata
+---@param include_children boolean
 ---@return Range6[]
 local function get_node_ranges(node, source, metadata, include_children)
   local range = vim.treesitter.get_range(node, source, metadata)
+  local child_count = node:named_child_count()
 
-  if include_children then
+  if include_children or child_count == 0 then
     return { range }
   end
 
@@ -529,7 +535,8 @@ local function get_node_ranges(node, source, metadata, include_children)
 
   local srow, scol, sbyte, erow, ecol, ebyte = Range.unpack6(range)
 
-  for i = 0, node:named_child_count() - 1 do
+  -- We are excluding children so we need to mask out their ranges
+  for i = 0, child_count - 1 do
     local child = node:named_child(i)
     local c_srow, c_scol, c_sbyte, c_erow, c_ecol, c_ebyte = child:range(true)
     if c_srow > srow or c_scol > scol then
@@ -561,7 +568,10 @@ end
 ---@param combined boolean
 ---@param ranges Range6[]
 local function add_injection(t, tree_index, pattern, lang, combined, ranges)
-  assert(type(lang) == 'string')
+  if #ranges == 0 then
+    -- Make sure not to add an empty range set as this is interpreted to mean the whole buffer.
+    return
+  end
 
   -- Each tree index should be isolated from the other nodes.
   if not t[tree_index] then
@@ -790,6 +800,20 @@ function LanguageTree:_edit(
     end
     return true
   end)
+
+  for _, child in pairs(self._children) do
+    child:_edit(
+      start_byte,
+      end_byte_old,
+      end_byte_new,
+      start_row,
+      start_col,
+      end_row_old,
+      end_col_old,
+      end_row_new,
+      end_col_new
+    )
+  end
 end
 
 ---@package
@@ -836,20 +860,17 @@ function LanguageTree:_on_bytes(
   )
 
   -- Edit trees together BEFORE emitting a bytes callback.
-  ---@private
-  self:for_each_child(function(child)
-    child:_edit(
-      start_byte,
-      start_byte + old_byte,
-      start_byte + new_byte,
-      start_row,
-      start_col,
-      start_row + old_row,
-      old_end_col,
-      start_row + new_row,
-      new_end_col
-    )
-  end, true)
+  self:_edit(
+    start_byte,
+    start_byte + old_byte,
+    start_byte + new_byte,
+    start_row,
+    start_col,
+    start_row + old_row,
+    old_end_col,
+    start_row + new_row,
+    new_end_col
+  )
 
   self:_do_callback(
     'bytes',
@@ -903,9 +924,9 @@ function LanguageTree:register_cbs(cbs, recursive)
   end
 
   if recursive then
-    self:for_each_child(function(child)
+    for _, child in pairs(self._children) do
       child:register_cbs(cbs, true)
-    end)
+    end
   end
 end
 
